@@ -12,7 +12,7 @@ mod tests;
 
 use crate::{
     bytecode::{encode_str, write_to_vec},
-    collections::key_map::{Key, KeyMap},
+    collections::key_map::{Handle, KeyMap},
     instruction::instruction_span,
     program::{CaoProgram, Label},
     Instruction, NodeId, VarName, VariableId,
@@ -43,20 +43,18 @@ pub struct Compiler<'a> {
     pub(crate) program: CaoProgram,
     next_var: RefCell<VariableId>,
 
-    /// maps lanes to their pre-hash-map keys
+    /// maps lanes to their metadata
     jump_table: KeyMap<LaneMeta>,
 
     locals: Box<arrayvec::ArrayVec<Local<'a>, 255>>,
     scope_depth: i32,
     current_card: i32,
     current_lane: LaneNode,
-
-    _m: std::marker::PhantomData<&'a ()>,
 }
 
 #[derive(Debug, Clone, Copy)]
 struct LaneMeta {
-    pub hash_key: Key,
+    pub hash_key: Handle,
     /// number of arguments
     pub arity: u32,
 }
@@ -94,7 +92,6 @@ impl<'a> Compiler<'a> {
             scope_depth: 0,
             current_lane: LaneNode::LaneId(0),
             current_card: -1,
-            _m: Default::default(),
         }
     }
 
@@ -143,11 +140,11 @@ impl<'a> Compiler<'a> {
         for (i, n) in compilation_unit.lanes.iter().enumerate() {
             self.current_lane = LaneNode::LaneId(i);
 
-            let indexkey = Key::from_u32(i as u32);
+            let indexkey = Handle::from_i64(i as i64);
             assert!(!self.jump_table.contains(indexkey));
             num_cards += n.cards.len();
 
-            let nodekey = Key::from_u32(
+            let nodekey = Handle::from_u32(
                 NodeId {
                     // we know that i fits in 16 bits from the check above
                     lane: i as u16,
@@ -162,7 +159,7 @@ impl<'a> Compiler<'a> {
             };
             self.jump_table.insert(indexkey, metadata).unwrap();
             if let Some(name) = n.name.as_ref() {
-                let namekey = Key::from_str(name.as_str()).expect("Failed to hash lane name");
+                let namekey = Handle::from_str(name.as_str()).expect("Failed to hash lane name");
                 if self.jump_table.contains(namekey) {
                     return Err(self.error(CompilationErrorPayload::DuplicateName(name.clone())));
                 }
@@ -200,7 +197,7 @@ impl<'a> Compiler<'a> {
                     lane: il as u16,
                     pos: 0,
                 };
-                let nodeid_hash = Key::from_u32(nodeid.into());
+                let nodeid_hash = Handle::from_u32(nodeid.into());
                 let handle = u32::try_from(self.program.bytecode.len())
                     .expect("bytecode length to fit into 32 bits");
                 self.program
@@ -278,7 +275,7 @@ impl<'a> Compiler<'a> {
                 self.error(CompilationErrorPayload::BadVariableName(param.to_string()))
             })?)?;
         }
-        for (ic, card) in cards.into_iter().enumerate() {
+        for (ic, card) in cards.iter().enumerate() {
             self.current_card = ic as i32;
             let nodeid = NodeId {
                 lane: il as u16,
@@ -292,7 +289,6 @@ impl<'a> Compiler<'a> {
     fn conditional_jump(
         &mut self,
         skip_instr: Instruction,
-        nodeid: NodeId,
         lane: &LaneNode,
     ) -> CompilationResult<()> {
         assert!(
@@ -307,34 +303,17 @@ impl<'a> Compiler<'a> {
         debug_assert_eq!(std::mem::size_of_val(&pos), 4);
         write_to_vec(pos, &mut self.program.bytecode);
         self.program.bytecode.push(Instruction::CallLane as u8);
-        self.encode_jump(nodeid, lane)?;
+        self.encode_jump(lane)?;
         Ok(())
     }
 
-    fn encode_jump(&mut self, nodeid: NodeId, lane: &LaneNode) -> CompilationResult<()> {
-        let to = match lane {
-            LaneNode::LaneName(lane) => self
-                .jump_table
-                .get(Key::from_str(lane).expect("Failed to hash jump target name"))
-                .ok_or_else(|| {
-                    self.error(CompilationErrorPayload::InvalidJump {
-                        src: nodeid,
-                        dst: lane.to_string(),
-                        msg: None,
-                    })
-                })?,
-            LaneNode::LaneId(id) => {
-                self.jump_table
-                    .get(Key::from_u32(*id as u32))
-                    .ok_or_else(|| {
-                        self.error(CompilationErrorPayload::InvalidJump {
-                            src: nodeid,
-                            dst: format!("Lane id {}", id),
-                            msg: None,
-                        })
-                    })?
-            }
-        };
+    fn encode_jump(&mut self, lane: &LaneNode) -> CompilationResult<()> {
+        let to = self.jump_table.get(Handle::from(lane)).ok_or_else(|| {
+            self.error(CompilationErrorPayload::InvalidJump {
+                dst: lane.clone(),
+                msg: None,
+            })
+        })?;
         write_to_vec(to.hash_key, &mut self.program.bytecode);
         write_to_vec(to.arity, &mut self.program.bytecode);
         Ok(())
@@ -360,7 +339,7 @@ impl<'a> Compiler<'a> {
     fn process_card(&mut self, nodeid: NodeId, card: &Card) -> CompilationResult<()> {
         let handle = u32::try_from(self.program.bytecode.len())
             .expect("bytecode length to fit into 32 bits");
-        let nodeid_hash = Key::from_u32(nodeid.into());
+        let nodeid_hash = Handle::from_u32(nodeid.into());
         self.program
             .labels
             .0
@@ -373,12 +352,19 @@ impl<'a> Compiler<'a> {
         }
         match card {
             Card::ForEach { variable, lane } => {
-                // TODO: assert that target lane has two arguments!
+                let target_lane = Handle::from(lane);
+                let arity = self.jump_table[target_lane].arity;
+                if arity != 2 {
+                    return Err(self.error(CompilationErrorPayload::InvalidJump {
+                        dst: lane.clone(),
+                        msg: Some("ForEach lanes need to have 2 parameters".to_string()),
+                    }));
+                }
                 self.read_var_card(variable)?;
                 self.program.bytecode.push(Instruction::BeginForEach as u8);
                 let block_begin = self.program.bytecode.len() as i32;
                 self.program.bytecode.push(Instruction::ForEach as u8);
-                self.encode_jump(nodeid, &lane)?;
+                self.encode_jump(lane)?;
                 // return to the repeat instruction
                 self.program.bytecode.push(Instruction::GotoIfTrue as u8);
                 write_to_vec(block_begin, &mut self.program.bytecode);
@@ -388,10 +374,18 @@ impl<'a> Compiler<'a> {
                 return Err(self.error(CompilationErrorPayload::Unimplemented("While cards")))
             }
             Card::Repeat(repeat) => {
+                let target_lane = Handle::from(repeat);
+                let arity = self.jump_table[target_lane].arity;
+                if arity != 1 {
+                    return Err(self.error(CompilationErrorPayload::InvalidJump {
+                        dst: repeat.clone(),
+                        msg: Some("Repeat lanes need to have 1 parameter".to_string()),
+                    }));
+                }
                 self.program.bytecode.push(Instruction::BeginRepeat as u8);
                 let block_begin = self.program.bytecode.len() as i32;
                 self.program.bytecode.push(Instruction::Repeat as u8);
-                self.encode_jump(nodeid, &repeat)?;
+                self.encode_jump(repeat)?;
                 // return to the repeat instruction
                 self.program.bytecode.push(Instruction::GotoIfTrue as u8);
                 write_to_vec(block_begin, &mut self.program.bytecode);
@@ -401,7 +395,7 @@ impl<'a> Compiler<'a> {
             }
             Card::SetVar(var) => {
                 let index = self.locals.len() as u32;
-                self.add_local(var.0)?;
+                self.add_local(*var.0)?;
                 self.program.bytecode.push(Instruction::SetLocalVar as u8);
                 write_to_vec(index, &mut self.program.bytecode);
             }
@@ -410,7 +404,7 @@ impl<'a> Compiler<'a> {
                 if variable.0.is_empty() {
                     return Err(self.error(CompilationErrorPayload::EmptyVariable));
                 }
-                let varhash = Key::from_bytes(variable.0.as_bytes());
+                let varhash = Handle::from_bytes(variable.0.as_bytes());
 
                 let id = self
                     .program
@@ -426,7 +420,7 @@ impl<'a> Compiler<'a> {
                     .variables
                     .names
                     .entry(*id)
-                    .or_insert_with(move || variable.0);
+                    .or_insert_with(move || *variable.0);
                 write_to_vec(*id, &mut self.program.bytecode);
             }
             Card::IfElse {
@@ -443,7 +437,7 @@ impl<'a> Compiler<'a> {
                 write_to_vec(pos, &mut self.program.bytecode);
                 // else
                 self.program.bytecode.push(Instruction::CallLane as u8);
-                self.encode_jump(nodeid, &else_lane)?;
+                self.encode_jump(else_lane)?;
 
                 self.program.bytecode.push(Instruction::Goto as u8);
                 let pos = instruction_span(Instruction::CallLane)
@@ -452,23 +446,23 @@ impl<'a> Compiler<'a> {
                 write_to_vec(pos, &mut self.program.bytecode);
                 // then
                 self.program.bytecode.push(Instruction::CallLane as u8);
-                self.encode_jump(nodeid, &then_lane)?;
+                self.encode_jump(then_lane)?;
             }
             Card::IfFalse(jmp) => {
                 // if the value is true we DON'T jump
-                self.conditional_jump(Instruction::GotoIfTrue, nodeid, &jmp)?;
+                self.conditional_jump(Instruction::GotoIfTrue, jmp)?;
             }
             Card::IfTrue(jmp) => {
                 // if the value is false we DON'T jump
-                self.conditional_jump(Instruction::GotoIfFalse, nodeid, &jmp)?;
+                self.conditional_jump(Instruction::GotoIfFalse, jmp)?;
             }
             Card::Jump(jmp) => {
-                self.encode_jump(nodeid, &jmp)?;
+                self.encode_jump(jmp)?;
             }
             Card::StringLiteral(c) => self.push_str(c.0.as_str()),
             Card::CallNative(c) => {
                 let name = &c.0;
-                let key = Key::from_str(name.as_str()).unwrap();
+                let key = Handle::from_str(name.as_str()).unwrap();
                 write_to_vec(key, &mut self.program.bytecode);
             }
             Card::ScalarInt(s) => {
@@ -509,7 +503,7 @@ impl<'a> Compiler<'a> {
         if scope < 0 {
             // global
             let mut next_var = self.next_var.borrow_mut();
-            let varhash = Key::from_bytes(variable.0.as_bytes());
+            let varhash = Handle::from_bytes(variable.0.as_bytes());
             let id = self
                 .program
                 .variables
@@ -524,7 +518,7 @@ impl<'a> Compiler<'a> {
                 .variables
                 .names
                 .entry(*id)
-                .or_insert_with(|| variable.0);
+                .or_insert_with(|| *variable.0);
             self.program.bytecode.push(Instruction::ReadGlobalVar as u8);
             write_to_vec(*id, &mut self.program.bytecode);
         } else {
