@@ -1,6 +1,3 @@
-mod fibonacci;
-
-use std::convert::TryInto;
 use test_env_log::test;
 
 use cao_lang::{
@@ -17,7 +14,7 @@ fn test_string_w_utf8() {
             .with_card(Card::SetGlobalVar(VarNode::from_str_unchecked("result")))],
     };
 
-    let program = compile(program, Some(CompileOptions::new())).expect("compile");
+    let program = compile(&program, Some(CompileOptions::new())).expect("compile");
 
     // Compilation was successful
 
@@ -25,12 +22,8 @@ fn test_string_w_utf8() {
     vm.run(&program).expect("run");
 
     let varid = program.variable_id("result").expect("varid");
-    let ptr = vm
-        .read_var(varid)
-        .expect("read var")
-        .try_into()
-        .expect("Expected a pointer");
-    let ress = unsafe { vm.get_str(ptr).expect("Failed to read string from vm") };
+    let ptr = vm.read_var(varid).expect("read var");
+    let ress = unsafe { ptr.as_str().expect("Failed to read string") };
 
     assert_eq!(test_str, ress);
 }
@@ -45,7 +38,7 @@ fn test_string_param() {
     }
 
     let fun = move |vm: &mut Vm<State>, arg: cao_lang::StrPointer| {
-        let vm_str = unsafe { vm.get_str(StrPointer(arg.0)).unwrap().to_string() };
+        let vm_str = unsafe { arg.get_str().unwrap().to_string() };
         vm.auxiliary_data.res = vm_str;
         Ok(())
     };
@@ -64,7 +57,7 @@ fn test_string_param() {
             ))))],
     };
 
-    let program = compile(cu, None).expect("compile");
+    let program = compile(&cu, None).expect("compile");
 
     vm.run(&program).expect("run");
     let aux = vm.unwrap_aux();
@@ -86,7 +79,7 @@ fn simple_if_statement() {
             ]),
         ],
     };
-    let program = compile(program, Some(CompileOptions::new())).expect("compile");
+    let program = compile(&program, Some(CompileOptions::new())).expect("compile");
 
     // Compilation was successful
 
@@ -111,7 +104,7 @@ fn simple_if_statement_skips_if_false() {
             ]),
         ],
     };
-    let program = compile(program, Some(CompileOptions::new())).unwrap();
+    let program = compile(&program, Some(CompileOptions::new())).unwrap();
 
     // Compilation was successful
 
@@ -145,7 +138,7 @@ fn if_else_test(condition: Card, true_res: Card, false_res: Card, expected_resul
             ]),
         ],
     };
-    let program = compile(program, Some(CompileOptions::new())).expect("compile");
+    let program = compile(&program, Some(CompileOptions::new())).expect("compile");
 
     // Compilation was successful
 
@@ -199,7 +192,7 @@ fn test_local_variable() {
             .with_card(Card::SetGlobalVar(VarNode::from_str_unchecked("bar")))],
     };
 
-    let program = compile(program, None).expect("compile");
+    let program = compile(&program, None).expect("compile");
 
     // Compilation was successful
 
@@ -227,7 +220,7 @@ fn local_variable_doesnt_leak_out_of_scope() {
         ],
     };
 
-    let program = compile(program, None).expect("compile");
+    let program = compile(&program, None).expect("compile");
 
     // Compilation was successful
 
@@ -257,7 +250,7 @@ fn simple_while_loop() {
         ],
     };
     /*let program =*/
-    match compile(program, Some(CompileOptions::new())).map_err(|e| e.payload) {
+    match compile(&program, Some(CompileOptions::new())).map_err(|e| e.payload) {
         Ok(_) => {
             panic!("Expected error, update this test pls")
         }
@@ -290,16 +283,19 @@ fn simple_for_loop() {
                 Card::ScalarInt(IntegerNode(5)),
                 Card::Repeat(LaneNode::LaneName("Loop".to_string())),
             ]),
-            Lane::default().with_name("Loop").with_cards(vec![
-                // Add 1 to the global 'result' variable in each iteration
-                Card::ScalarInt(IntegerNode(1)),
-                Card::ReadVar(VarNode::from_str_unchecked("result")),
-                Card::Add,
-                Card::SetGlobalVar(VarNode::from_str_unchecked("result")),
-            ]),
+            Lane::default()
+                .with_name("Loop")
+                .with_arg("i")
+                .with_cards(vec![
+                    // Add i to the global 'result' variable in each iteration
+                    Card::ReadVar(VarNode::from_str_unchecked("i")),
+                    Card::ReadVar(VarNode::from_str_unchecked("result")),
+                    Card::Add,
+                    Card::SetGlobalVar(VarNode::from_str_unchecked("result")),
+                ]),
         ],
     };
-    let program = compile(program, Some(CompileOptions::new())).expect("compile");
+    let program = compile(&program, Some(CompileOptions::new())).expect("compile");
 
     // Compilation was successful
 
@@ -309,7 +305,7 @@ fn simple_for_loop() {
     let res = vm
         .read_var_by_name("result", &program.variables)
         .expect("Failed to read result variable");
-    assert_eq!(res, Value::Integer(5));
+    assert_eq!(res, Value::Integer((0..5).fold(0, std::ops::Add::add)));
 }
 
 #[test]
@@ -323,7 +319,7 @@ fn call_native_test() {
             )))])],
     };
 
-    let prog = compile(cu, CompileOptions::new()).unwrap();
+    let prog = compile(&cu, CompileOptions::new()).unwrap();
 
     struct State {
         called: bool,
@@ -414,7 +410,7 @@ lanes:
           val: "func3"
 "#;
     let cu = serde_yaml::from_str(PROG).unwrap();
-    let prog = compile(cu, CompileOptions::new()).unwrap();
+    let prog = compile(&cu, CompileOptions::new()).unwrap();
 
     vm.run(&prog).expect("run failed");
 
@@ -444,7 +440,7 @@ fn jump_lane_w_params_test() {
         ],
     };
 
-    let program = compile(cu, CompileOptions::new()).expect("compile");
+    let program = compile(&cu, CompileOptions::new()).expect("compile");
 
     let mut vm = Vm::new(()).unwrap();
     vm.run(&program).expect("run");
@@ -473,7 +469,7 @@ fn len_test_empty() {
             .with_card(Card::SetGlobalVar(VarNode::from_str_unchecked("g_result")))],
     };
 
-    let program = compile(cu, CompileOptions::new()).expect("compile");
+    let program = compile(&cu, CompileOptions::new()).expect("compile");
 
     let mut vm = Vm::new(()).unwrap();
     vm.run(&program).expect("run");
@@ -493,19 +489,19 @@ fn len_test_happy() {
         lanes: vec![Lane::default()
             .with_name("main")
             .with_card(Card::CreateTable)
-            .with_card(Card::SetVar(t))
+            .with_card(Card::SetVar(t.clone()))
             // first property
-            .with_card(Card::ReadVar(t))
+            .with_card(Card::ReadVar(t.clone()))
             .with_card(Card::StringLiteral(StringNode("asd".to_string())))
             .with_card(Card::ScalarInt(IntegerNode(42)))
             .with_card(Card::SetProperty)
             // same property as above
-            .with_card(Card::ReadVar(t))
+            .with_card(Card::ReadVar(t.clone()))
             .with_card(Card::StringLiteral(StringNode("asd".to_string())))
             .with_card(Card::ScalarInt(IntegerNode(42)))
             .with_card(Card::SetProperty)
             // new property
-            .with_card(Card::ReadVar(t))
+            .with_card(Card::ReadVar(t.clone()))
             .with_card(Card::StringLiteral(StringNode("basdasd".to_string())))
             .with_card(Card::ScalarInt(IntegerNode(42)))
             .with_card(Card::SetProperty)
@@ -513,7 +509,7 @@ fn len_test_happy() {
             .with_card(Card::SetGlobalVar(VarNode::from_str_unchecked("g_result")))],
     };
 
-    let program = compile(cu, CompileOptions::new()).expect("compile");
+    let program = compile(&cu, CompileOptions::new()).expect("compile");
 
     let mut vm = Vm::new(()).unwrap();
     vm.run(&program).expect("run");
@@ -524,78 +520,3 @@ fn len_test_happy() {
 
     assert_eq!(len, Value::Integer(2));
 }
-
-#[test]
-fn for_each_1() {
-    let cu: CaoIr = serde_yaml::from_str(
-        r#"
-lanes: 
-    - name: main
-      cards:
-        # --- init the example table ---
-        - ty: CreateTable
-        - ty: SetVar
-          val: t
-        - ty: ReadVar
-          val: t
-        - ty: StringLiteral
-          val: f1
-        - ty: ScalarInt
-          val: 1
-        - ty: SetProperty
-        - ty: ReadVar
-          val: t
-        - ty: StringLiteral
-          val: f2
-        - ty: ScalarInt
-          val: 2
-        - ty: SetProperty
-        - ty: ReadVar
-          val: t
-        - ty: StringLiteral
-          val: f3
-        - ty: ScalarInt
-          val: 3
-        - ty: SetProperty
-        - ty: ScalarInt
-          val: 0
-        - ty: SetGlobalVar
-          val: g_result
-        # --- init done ---
-        - ty: ForEach
-          val:
-            lane: 
-              LaneName: pog
-            variable: t
-
-
-    - name: pog
-      arguments:
-        - "key"
-        - "table"
-      cards:
-        - ty: ReadVar
-          val: table
-        - ty: ReadVar
-          val: key
-        - ty: GetProperty
-        - ty: ReadVar
-          val: g_result
-        - ty: Add
-        - ty: SetGlobalVar
-          val: g_result
-        "#,
-    )
-    .unwrap();
-
-    let program = compile(cu, CompileOptions::new()).expect("compile");
-
-    let mut vm = Vm::new(()).unwrap();
-    vm.run(&program).expect("run");
-
-    let res = vm
-        .read_var_by_name("g_result", &program.variables)
-        .expect("Failed to read result variable");
-
-    assert_eq!(res, Value::Integer(6));
-}
